#version 300 es

#define TEETH_WHITENING
#define teethWhiteningCoeff 1.0
#define SOFT_LIGHT_LAYER
#define NORMAL_LAYER
#define SOFT_SKIN
#define skinSoftIntensity 1.0
#define SHARPEN_TEETH
#define teethSharpenIntensity 0.2
#define SHARPEN_EYES
#define eyesSharpenIntensity 0.3
#define PSI 0.1

#define SOFT_LIGHT_LAYER_HALF

precision mediump float;

layout( location = 0 ) out vec4 F;

in vec3 maskColor;
in vec2 var_uv, var_bg_uv;

uniform sampler2D lookupTexEyes, lookupTexTeeth, glfx_BACKGROUND;

#if defined(EYES_HIGHLIGHT)
uniform sampler2D tex_highlight;
#endif
#ifdef SOFT_LIGHT_LAYER
uniform sampler2D tex_softLight;
#endif
#ifdef NORMAL_LAYER
uniform sampler2D  tex_normalMakeup;
#endif

 layout(std140) uniform glfx_OCCLUSION_DATA
 {
    highp vec4 glfx_OCCLUSION_RECT;
    highp vec4 glfx_SCREEN;
 };

vec3 textureLookup(vec3 originalColor, sampler2D lookupTexture)
{
    const float epsilon = 0.000001;
    const float lutSize = 512.0;

    float blueValue = (originalColor.b * 255.0) / 4.0;

    vec2 mulB = clamp(floor(blueValue) + vec2(0.0, 1.0), 0.0, 63.0);
    vec2 row = floor(mulB / 8.0 + epsilon);
    vec4 row_col = vec4(row, mulB - row * 8.0);
    vec4 lookup = originalColor.ggrr * (63.0 / lutSize) + row_col * (64.0 / lutSize) + (0.5 / lutSize);

    float factor = blueValue - mulB.x;

    vec3 sampled1 = textureLod(lookupTexture, lookup.zx, 0.).rgb;
    vec3 sampled2 = textureLod(lookupTexture, lookup.wy, 0.).rgb;

    vec3 res = mix(sampled1, sampled2, factor);
    return res;
}

vec3 whitening(vec3 originalColor, float factor, sampler2D lookup) {
    vec3 color = textureLookup(originalColor, lookup);
    return mix(originalColor, color, factor);
}

vec3 sharpen(vec3 originalColor, float factor) {
    float dx = glfx_SCREEN.z;
    float dy = glfx_SCREEN.w;
    
    vec3 total = 5.0 * originalColor
    - texture(glfx_BACKGROUND, vec2(var_bg_uv.x-dx, var_bg_uv.y-dy)).xyz
    - texture(glfx_BACKGROUND, vec2(var_bg_uv.x+dx, var_bg_uv.y-dy)).xyz
    - texture(glfx_BACKGROUND, vec2(var_bg_uv.x-dx, var_bg_uv.y+dy)).xyz
    - texture(glfx_BACKGROUND, vec2(var_bg_uv.x+dx, var_bg_uv.y+dy)).xyz;

    vec3 result = mix(originalColor, total, factor);
    return clamp(result, 0.0, 1.0);
}

vec3 softSkin(vec3 originalColor, float factor) {
    vec3 screenColor = originalColor;

    float dx = 4.0 / glfx_SCREEN.x;
    float dy = 4.0 / glfx_SCREEN.y;
    
    vec3 nextColor0 = texture(glfx_BACKGROUND, vec2(var_bg_uv.x-dx, var_bg_uv.y-dy)).xyz;
    vec3 nextColor1 = texture(glfx_BACKGROUND, vec2(var_bg_uv.x+dx, var_bg_uv.y-dy)).xyz;
    vec3 nextColor2 = texture(glfx_BACKGROUND, vec2(var_bg_uv.x-dx, var_bg_uv.y+dy)).xyz;
    vec3 nextColor3 = texture(glfx_BACKGROUND, vec2(var_bg_uv.x+dx, var_bg_uv.y+dy)).xyz;
    
    float intensity = screenColor.g;
    vec4 nextIntensity = vec4(nextColor0.g, nextColor1.g, nextColor2.g, nextColor3.g);
    

    vec4 lg = nextIntensity - intensity;
    

    vec4 curr = max(0.367 - abs(lg * (0.367*0.6/(1.41*PSI))), 0.);
    
    float summ = 1.0 + curr.x + curr.y + curr.z + curr.w;
    screenColor += (nextColor0 * curr.x + nextColor1 * curr.y + nextColor2 * curr.z + nextColor3 * curr.w);
    screenColor = screenColor * (factor / summ);
    
    screenColor = originalColor*(1.-factor) + screenColor;
    return screenColor;
}

float blendSoftLight(float a, float b) {
    return (a+2.*b*(1.-a))*a;
}

vec3 blendSoftLight(vec3 base, vec3 blend) {
    return vec3(blendSoftLight(base.r,blend.r),blendSoftLight(base.g,blend.g),blendSoftLight(base.b,blend.b));
}

vec3 blendSoftLight(vec3 base, vec3 blend, float opacity) {
    return (blendSoftLight(base, blend) * opacity + base * (1.0 - opacity));
}

void main()
{
    vec3 res = texture( glfx_BACKGROUND, var_bg_uv ).xyz;
    
#ifdef SOFT_SKIN
    res = softSkin(res, maskColor.r * skinSoftIntensity);
#endif
    
#ifdef SHARPEN_TEETH
    res = sharpen(res, maskColor.g * teethSharpenIntensity);
#endif
    
#if defined(TEETH_WHITENING)
    res = whitening(res, maskColor.g * teethWhiteningCoeff, lookupTexTeeth);
#endif
    
    
#ifdef SHARPEN_EYES
    res = sharpen(res, maskColor.b * eyesSharpenIntensity);
#endif
    
#if defined(EYES_WHITENING)
    res = whitening(res, maskColor.b * eyesWhiteningCoeff, lookupTexEyes);
#endif
    
#if defined(EYES_HIGHLIGHT)
    vec2 uv_high = var_uv; 

    #ifdef EYES_HIGHLIGHT_HALF
        uv_high.x = abs(2.0 * (uv_high.x - 0.5)); 
    #endif

    res = res + vec4( texture( tex_highlight, uv_high ).xyz, 0. );
#endif

#ifdef SOFT_LIGHT_LAYER
    vec2 uv_soft = var_uv; 

    #ifdef SOFT_LIGHT_LAYER_HALF
        uv_soft.x = abs(2.0 * (uv_soft.x - 0.5)); 
    #endif

    res.xyz = blendSoftLight( res.xyz, texture( tex_softLight, uv_soft ).xyz );
#endif

#ifdef NORMAL_LAYER
    vec2 uv_norm = var_uv; 

    #ifdef NORMAL_LAYER_HALF
        uv_norm.x = abs(2.0 * (uv_norm.x - 0.5)); 
    #endif

    vec4 makeup2 = texture( tex_normalMakeup, uv_norm );
    res.xyz = mix( res.xyz, makeup2.xyz, makeup2.w );
#endif

    F = vec4(res,1.);
}